#!/bin/sh
#------------------------------------------------------------------------------
# httpDNS - fli4l Web Interface                                           1.2.0
#   Copyright (C) 2006-2008 Tiger
#
#   This OPT package is licensed under Free Software Foundation, Inc.,
#   General Public License Version 2, June 1991
#------------------------------------------------------------------------------

# NAMING CONVENTIONS
#  User defined globals are in capital letters with preceeding "HTTPDNS_", e.g. HTTPDNS_SERVER_N.
#  Program scope globals are named "httpdns*" to avoid conflicts with other programs running, e.g. httpdnsservername.
#  Routines (except global routines) are named hierarchically, e.g. show_DNSTables_List_Output() = "Output" called by "List" called by "DNSTables".
#  Routines starting with "get_" return the appropriate value.
#  Routines starting with "check_" return "yes" if successful.
#  Routines starting with "show_" mainly output to screen.
#  Routine parameters are ref'ed at the beginning of the routine, e.g. servername=$1.

. /srv/www/include/cgi-helper

HTTPDNS_PACKAGE="httpDNS"
HTTPDNS_VERSION="1.2.0"
# File System Globals
: ${HTTPDNS_TEMP_FOLDER:="/tmp"}
: ${HTTPDNS_TEMP:="$HTTPDNS_TEMP_FOLDER/httpdns.tmp"}
: ${HTTPDNS_DEBUG_REQUEST:="$HTTPDNS_TEMP_FOLDER/httpdns-request.tmp"}
: ${HTTPDNS_DEBUG_RESPONSE:="$HTTPDNS_TEMP_FOLDER/httpdns-response.tmp"}
: ${HTTPDNS_FOLDERCHAR:="~"}
: ${HTTPDNS_SERVERFILE:="httpdns.php"}
: ${HTTPDNS_SERVER:="/srv/www/admin/$HTTPDNS_SERVERFILE"}
: ${HTTPDNS_SERVERTEMP:="$HTTPDNS_TEMP_FOLDER/$HTTPDNS_SERVERFILE"}
: ${HTTPDNS_FTPPROGRAM:="/usr/bin/ftp"}
: ${HTTPDNS_NETCAT:="/usr/bin/netcat -w 30"}
: ${HTTPDNS_WGET:="/usr/bin/wget --timeout=30"}
: ${HTTPDNS_DNSFILEEXTENSION:=".dns"}
: ${HTTPDNS_HOSTSFILE:="/etc/hosts.httpdns"}
: ${HTTPDNS_DNSMASQPID:="/var/run/dnsmasq.pid"}
# Connection Globals
: ${HTTPDNS_QUERYPORT:=80}
: ${HTTPDNS_IP_SERVER:=1}
# Defaults
: ${HTTPDNS_LOGLEVEL:=5}
: ${HTTPDNS_SERVER_DEFAULT_DNSFILE:="httpdns.dns"}
: ${HTTPDNS_SERVER_DEFAULT_LOGLEVEL:=5}
: ${HTTPDNS_SERVER_DEFAULT_LIFETIME:=0}
# Various Globals
: ${HTTPDNS_RIGHTS_SECTION:="httpdns"}

# Program Globals
httpdnsheaderline=""
httpdnsverbose=""    # yes=output messages to console, no=do not output anything, empty=output HTML
httpdnserrorlevel=0

log()
{
  message=$1
  # ------------------------
  if [ -n "$HTTPDNS_LOGFILE" -a 0$HTTPDNS_LOGLEVEL -ge 1 ]
  then
    if [ -n "`echo $message | grep \"ERROR\"`" -o "`echo $message | cut -c 1`" != " " -o 0$HTTPDNS_LOGLEVEL -ge 5 ]
    then
      echo "$(date | cut -c 5-19) $HOSTNAME httpdns: $message" >> $HTTPDNS_LOGFILE
    fi
  fi
  if [ "$httpdnsverbose" = "yes" ]; then echo "$message"; fi
}

errorlevel()
{
  httpdnserrorlevel==$1
  # ------------------------
}

check_UserRight()
{
  required_right=$1
  # ------------------------
  rights=""
  if [ -n "$REMOTE_USER" ]
  then
    rights="`grep ^$REMOTE_USER= /etc/httpd/rights | cut -d= -f2`"
    if [ -n "$rights" -a -n "`echo $rights | grep \"$HTTPDNS_RIGHTS_SECTION:\"`" ]
    then
      rights="`echo $rights | sed -n 's/.*'$HTTPDNS_RIGHTS_SECTION':\([^ ]*\).*$/\1/p'`"
    else
      rights="all"
    fi
  fi
  if [ -n "`echo $rights | grep $required_right`" -o -n "`echo $rights | grep all`" ]
  then
    echo "yes"
  else
    errorlevel 1
  fi
}

check_ServerVersion()
{
  requiredversion=$1
  serverversion=$2
  # ------------------------
  req_version=`echo $requiredversion | sed -n 's|\([0-9]*\)\.\([0-9]*\)\.\([0-9]*\)$|\1|p'`
  req_release=`echo $requiredversion | sed -n 's|\([0-9]*\)\.\([0-9]*\)\.\([0-9]*\)$|\2|p'`
  req_build=`echo $requiredversion | sed -n 's|\([0-9]*\)\.\([0-9]*\)\.\([0-9]*\)$|\3|p'`
  chk_version=`echo $serverversion | sed -n 's|\([0-9]*\)\.\([0-9]*\)\.\([0-9]*\)$|\1|p'`
  chk_release=`echo $serverversion | sed -n 's|\([0-9]*\)\.\([0-9]*\)\.\([0-9]*\)$|\2|p'`
  chk_build=`echo $serverversion | sed -n 's|\([0-9]*\)\.\([0-9]*\)\.\([0-9]*\)$|\3|p'`
  if [ $chk_version -gt $req_version -o \( $chk_version -eq $req_version -a \( $chk_release -gt $req_release -o \( $chk_release -eq $req_release -a $chk_build -ge $req_build \) \) \) ]
  then
    echo "yes"
  fi
}

check_IfWeAreOnline()
{
  if [ -n "$HTTPDNS_INTERFACE" ]
  then
    if [ -n "`get_LocalIP`" ]
    then
      echo "yes"
    fi
  else
    if [ -n "`ls -1 /var/run/*.ip 2>nul`" ]
    then
      echo "yes"
    fi
  fi
}

get_FastRandom()
{
  echo $(head -n 3 /dev/urandom | sed -n 's/[^0-9]//gp' | tr "\n" "0") | sed -n 's/\(.\{1,7\}\).*/0\1/p'
}

get_DummyIP()
{
  for c in `seq 1 4`; do echo `expr $(get_FastRandom) % 256`; done | tr '\n' '.' | sed 's/\.$//'
}

get_LocalIP()
{
  if [ -n "$HTTPDNS_INTERFACE" ]
  then
    if [ -e "/sbin/ifconfig" ]
    then
      ifconfig $HTTPDNS_INTERFACE | sed -n 's/^ *inet addr:\([0-9.]*\).*$/\1/p'
    else
      ip addr show $HTTPDNS_INTERFACE | sed -n 's/^ *inet \([0-9.]*\).*$/\1/p'
    fi
  fi
}

get_ServernameFromFilename()
{
  filename=$1
  # ------------------------
  echo $filename | sed 's|.*/||;s|'$HTTPDNS_FOLDERCHAR'|/|g;s|\(.*\)'$HTTPDNS_DNSFILEEXTENSION'$|\1|;s|//|://|'
}

get_FilenameFromServername()
{
  servername=$1
  # ------------------------
  echo $servername | sed 's|/|'$HTTPDNS_FOLDERCHAR'|g;s|:||g;s|.*|'$HTTPDNS_DNSFOLDER'/&'$HTTPDNS_DNSFILEEXTENSION'|'
}

get_FileName()
{
  filenameandpath=$1
  # ------------------------
  echo $filenameandpath | sed 's,^\(.*\)/\(.*\)$,\2,'
}

get_Path()
{
  filenameandpath=$1
  # ------------------------
  echo $filenameandpath | sed -n 's,^\(.*\)/\(.*\)$,\1,p'
}

get_HTTPDNS_SERVER_()
{
  serverindex=$1
  variable=$2
  # ------------------------
  eval result="\$HTTPDNS_SERVER_"$serverindex$variable
  echo "$result"
}

get_Servername()
{
  serverindex=$1
  # ------------------------
  echo "`get_HTTPDNS_SERVER_ $serverindex _NAME | sed 's|/$||;s|.*|http://&|;s|http://\(http[s]*://.*\)$|\1|'`"
}

get_Encrypted()
{
  ipaddress=$1
  hostname=$2
  key=$3
  magic=$4
  # ------------------------
  encrypted=""
  if [ -n "$ipaddress" -a -n "$hostname" -a -n "$key" ]
  then
    # Convert IP address from dec to hex, add leading zero to single digit numbers, concat to string
    for ip_part in `echo $ipaddress | sed -n 's/\./ /gp'`
    do
      nibble_high=`expr $ip_part / 16`
      nibble_low=`expr $ip_part % 16`
      encrypted=$encrypted"`expr substr 123456789abcdef $nibble_high 1 | sed -n 's/.*/0&/;s/.*\(.\)/\1/p'`"
      encrypted=$encrypted"`expr substr 123456789abcdef $nibble_low 1 | sed -n 's/.*/0&/;s/.*\(.\)/\1/p'`"
    done
    encrypted="`echo $encrypted$hostname | tr '[A-Z]' '[a-z]'`"
    encrypted="$encrypted`echo $encrypted | md5sum | sed -n 's/^\(.\{32\}\).*$/\1/p'`"
    if [ -n "$magic" ]
    then
      RANDOM=`expr \`date +'%s %% 32676'\``
      crc_length=32
      str_offset=`expr length $encrypted - $crc_length`
      for magic_char in `echo $magic | sed -n 's/./& /gp'`
      do
        RANDOM=$RANDOM
        crc_length=`expr $crc_length + 1`
        char_pos=`expr $RANDOM % $crc_length + $str_offset`
        encrypted=`echo $encrypted | cut -c-$char_pos`$magic_char`echo $encrypted | cut -c\`expr $char_pos + 1\`-`
      done
    fi
    encrypted="`echo $encrypted | tr '0-9a-z@.|_!' \"$key\"`"
  fi
  echo $encrypted
}

get_Decrypted()
{
  encrypted=$1
  key=$2
  # ------------------------
  record=""
  if [ -n "$encrypted" -a -n "$key" ]
  then
    decrypted="`echo $encrypted | tr \"$key\" '0-9a-z@.|_!'`"
    if [ "`echo $decrypted | sed -n 's/^\(.*\).\{32\}$/\1/p' | md5sum | sed -n 's/^\(.\{32\}\).*$/\1/p'`" = "`echo $decrypted | sed -n 's/^.*\(.\{32\}\)$/\1/p'`" ]
    then
      value=""
      ipaddress=""
      for ip_part in `echo $decrypted |  sed -n 's/^\(.\{8\}\).*/\1/;s/./& /gp'`
      do
        if [ -z "$value" ]
        then
          value=`expr index 123456789abcdef $ip_part`
        else
          ipaddress=$ipaddress"."`expr 16 \* $value + index 123456789abcdef $ip_part`
          value=""
        fi
      done
      ipaddress=`echo $ipaddress | cut -c2-`
      hostname="`echo $decrypted | sed -n 's/^\(.\{8\}\)\(.*\)\(.\{14\}\)\(.\{32\}\)$/\2/p'`"
      timestamp="`echo $decrypted | sed -n 's/^\(.\{8\}\)\(.*\)\(.\{14\}\)\(.\{32\}\)$/\3/p'`"
      record="$hostname $timestamp $ipaddress"
    fi
  fi
  echo "$record"
}

get_DecryptedFile()
{
  filename=$1
  key=$2
  # ------------------------
  if [ -f $filename ]
  then
    cat $filename | while read line; do get_Decrypted "$line" "$key"; done | sed '/^$/d' | sort
  fi
}

get_LocalizedTime()
{
  timezone=$1
  timestamp=$2
  # ------------------------
  ts_date=`echo $timestamp | cut -c 1-8`        # cut off time
  ts_time=`echo $timestamp | cut -c 9-12`       # crop hours and minutes
  ts_time=`expr 2$ts_time $timezone`     # apply timezone
  if [ "`echo \"$timezone\" | cut -c 2`" = "-" ]
  then
    if [ `echo $ts_time | cut -c 4-5` -gt 59 ]  # correct minutes
    then
      ts_time=`expr $ts_time - 40`
    fi
    if [ `echo $ts_time | cut -c 2-3` -gt 23 ]  # correct hours
    then
      ts_time=`expr $ts_time - 7600`
      ts_date=`expr $ts_date - 1`
    fi
  fi
  ts_time=$ts_date"`echo $ts_time | sed -n 's/.*\(.\{4\}\)$/\1/p'`""`echo $timestamp | cut -c 13-14`" # concat date and time
  ts_time=`echo $ts_time | sed -n 's/^\(.\{4\}\)\(.\{2\}\)\(.\{2\}\)\(.\{2\}\)\(.\{2\}\)\(.\{2\}\)$/\2\3\4\5\1.\6/p'`
  date -d $ts_time '+%Y-%m-%d %H:%M:%S'
}

get_ServerType()
{
  echo "$httpdnsheaderline" | sed -n 's/.*'$HTTPDNS_PACKAGE' Server (\(.*\)) Version \([0-9.]*\).*$/\1/p'
}

get_ServerVersion()
{
  echo "$httpdnsheaderline" | sed -n 's/.*'$HTTPDNS_PACKAGE' Server (\(.*\)) Version \([0-9.]*\).*$/\2/p'
}

get_ServerResponse()
{
  echo "$httpdnsheaderline" | sed -n 's/.*'$HTTPDNS_PACKAGE' Server (\(.*\)) Version \([0-9.]*\) : \(.*\)$/\3/p'
}

erase_file()
{
  filename=$1
  # ------------------------
  if [ -n "$filename" ]
  then
    if [ -f $filename ]
    then
      rm $filename
    fi
  fi
}

show()
{
  text=$*
  # ------------------------
  if [ -z "$httpdnsverbose" ]; then echo "$text"; fi
}

show_LocalIP()
{
  show "`show_tab_header \"$_httpdns_table_currentip\" no`"
    show "<table class=\"normtable\">"
      show "<tr>"
        if [ -n "`check_IfWeAreOnline`" ]
        then
          if [ -n "$HTTPDNS_INTERFACE" ]
          then
              show "<th>$_httpdns_header_device"
              show "<th>$_httpdns_header_ipaddress</th>"
            show "</tr>"
            show "<tr>"
              show "<td>$HTTPDNS_INTERFACE</td>"
              show "<td>`get_LocalIP`</td>"
          else
              show "<th>$_httpdns_header_source"
              show "<th>$_httpdns_header_ipaddress</th>"
            show "</tr>"
            show "<tr>"
              if [ 0$HTTPDNS_IP_SERVER -le $HTTPDNS_SERVER_N ]
              then
                query_Server $HTTPDNS_IP_SERVER "getlocalip"
              else
                show "<td align=center colspan=2 class=\"httpdns_fail\">$_httpdns_status_noaccess</td>"
              fi
          fi
        else
          show "<td align=center colspan=2 class=\"httpdns_fail\">$_httpdns_status_offline</td>"
        fi
        if [ $HTTPDNS_SERVER_N -gt 0 ]
        then
          if [ -n "`check_UserRight sync`" ]
          then
            show "<td><a href=\"$myname?action=query\"><img src=\"../img/httpdns_syncall.gif\" border=0 title=\"$_httpdns_button_syncall\"></a></td>"
          fi
        fi
      show "</tr>"
    show "</table>"
  show "`show_tab_footer`"
}

show_DNSTables()
{
  serversshown=""
  if [ $HTTPDNS_SERVER_N -gt 0 ]
  then
    for serverindex in `seq 1 $HTTPDNS_SERVER_N`
    do
      servername="`get_Servername $serverindex`"
      show_DNSTables_List "$servername" $serverindex
      serversshown=$serversshown"~$servername~\n"
    done
  fi
  if [ -d $HTTPDNS_DNSFOLDER ]
  then
    for dnsfile in $(ls -1 $HTTPDNS_DNSFOLDER/*$HTTPDNS_DNSFILEEXTENSION)
    do
      servername="`get_ServernameFromFilename \"$dnsfile\"`"
      if [ -z "`echo \"$serversshown\" | grep \"~$servername~\"`" ]
      then
        show_DNSTables_List "$servername"
      fi
    done
  fi
}

show_DNSTables_List()
{
  servername=$1
  serverindex=$2
  # ------------------------
  if [ -n "$servername" ]
  then
    show "<br>"
    show "`show_tab_header \"$servername\" no`"
    filename="`get_FilenameFromServername \"$servername\"`"
    show "<table class=\"normtable\">"
      if [ -f $filename ]
      then
        show "<tr>"
          show "<th colspan=2>$_httpdns_header_synced</th><td>`date -r $filename '+%Y-%m-%d %H:%M:%S'`</td>"
          if [ -n "$serverindex" ]
          then
            if [ -n "`check_UserRight sync`" ]
            then
              show "<td rowspan=2><a href=\"$myname?action=query&server=$serverindex\"><img src=\"../img/httpdns_sync.gif\" border=0 title=\"$_httpdns_button_sync\"></a></td>"
            fi
          fi
        show "</tr>"
        show "<tr>"
          show "<th>$_httpdns_header_name</th>"
          show "<th>$_httpdns_header_ipaddress</th>"
          show "<th>$_httpdns_header_timestamp</th>"
        show "</tr>"
        timezone="`date +%z | sed -n 's/[+-]/ & /p'`"
        cat $filename | while read hostname timestamp ipaddress
        do
          show "<tr>"
            show "<td>$hostname</td><td>$ipaddress</td><td>`get_LocalizedTime \"$timezone\" \"$timestamp\"`</td>"
          show "</tr>"
          [ 0 -eq 1 ]
        done && show "<tr><td colspan=3><b>$_httpdns_status_nodata</b></td></tr>"
      else
        show "<tr>"
          show "<td class=\"httpdns_fail\">$_httpdns_info_notsynced</td>"
          if [ -n "$serverindex" ]
          then
            if [ -n "`check_UserRight sync`" ]
            then
              show "<td><a href=\"$myname?action=query&server=$serverindex\"><img src=\"../img/httpdns_sync.gif\" border=0 title=\"$_httpdns_button_sync\"></a></td>"
            fi
          fi
        show "</tr>"
      fi
    show "</table>"
    show "`show_tab_footer`"
  fi
}

show_AdminTable()
{
  if [ $HTTPDNS_SERVER_N -gt 0 ]
  then
    show "<table class=\"normtable\">"
      show "<tr>"
        show "<th>$_httpdns_header_server</th>"
        show "<th>$_httpdns_header_version</th>"
        show "<th>$_httpdns_header_command</th>"
      show "</tr>"
      for serverindex in `seq 1 $HTTPDNS_SERVER_N`
      do
        show "<tr>"
          query_Server $serverindex "version"
          show "<td>"
            show "<a href=\"$myname?action=viewlog&server=$serverindex\"><img src=\"../img/httpdns_viewlog.gif\" border=0 title=\"$_httpdns_button_viewlog\"></a>"
            show "<a href=\"$myname?action=clearlog&server=$serverindex\"><img src=\"../img/httpdns_clearlog.gif\" border=0 title=\"$_httpdns_button_clearlog\"></a>"
            if [ -n "`get_HTTPDNS_SERVER_ $serverindex _FTPURL`" -a -n "`get_HTTPDNS_SERVER_ $serverindex _FTPUSER`" -a -n "`get_HTTPDNS_SERVER_ $serverindex _FTPPWD`" ]
            then
              show "<a href=\"$myname?action=upload&server=$serverindex\"><img src=\"../img/httpdns_upload.gif\" border=0 title=\"$_httpdns_button_upload\"></a>"
            else
              show "<a href=\"$myname?action=manual&server=$serverindex\"><img src=\"../img/httpdns_manual.gif\" border=0 title=\"$_httpdns_button_manual\"></a>"
            fi
            if [ -n "`check_ServerVersion 1.1.1 $(get_ServerVersion)`" ]
            then
              show "<a href=\"$myname?action=linkgen&server=$serverindex\"><img src=\"../img/httpdns_linkgen.gif\" border=0 title=\"$_httpdns_button_linkgen\"></a>"
            fi
            if [ -n "`check_ServerVersion 1.1.7 $(get_ServerVersion)`" ]
            then
              show "<a href=\"$myname?action=killhost&server=$serverindex\"><img src=\"../img/httpdns_killhost.gif\" border=0 title=\"$_httpdns_button_killhost\"></a>"
            fi
          show "</td>"
        show "</tr>"
      done
    show "</table>"
  else
    show "`show_error \"$HTTPDNS_PACKAGE\" \"$_httpdns_error_noserver\"`"
  fi
}

show_ServerLog()
{
  serverindex=$1
  command=$2
  # ------------------------
  if [ -n "$serverindex" ]
  then
    case $command in
      "viewlog")
        show "`show_tab_header \"$_httpdns_table_viewlog\" no`"
      ;;
      "clearlog")
        show "`show_tab_header \"$_httpdns_table_clearlog\" no`"
      ;;
      *)
        show "`show_tab_header \"\" no`"
      ;;
    esac
      show "<table class=\"normtable\">"
        show "<tr>"
          show "<th>$_httpdns_header_server</th>"
          show "<th>$_httpdns_header_status</th>"
        show "</tr>"
        show "<tr>"
          query_Server $serverindex $command
        show "</tr>"
      show "</table>"
    show "`show_tab_footer`"
  fi
}

show_RandomKeys()
{
  show "<h2>$_httpdns_info_keygen</h2>"
  show "<table class=\"normtable\">"
    RANDOM=$(expr \( `date +%S` + 1 \) \* \( `date +%M` + 1 \) \* \( `date +%H` + 1 \) \* `date +%d` % 32768)
    for key_index in `seq 1 3`
    do
      ref_index=1
      key=""
      keyset="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890@.|_!"
      for char_index in `seq 1 41`
      do
        RANDOM=`expr \( $RANDOM + $key_index \* $ref_index \* $char_index \) % 32768`
        ref_index=`expr $RANDOM % $(expr length "$keyset") + 1`
        key="$key`echo $keyset | cut -c $ref_index`"
        if [ $ref_index -gt 1 ]
        then
          keyset="`echo $keyset | cut -c -$(expr $ref_index - 1),$(expr $ref_index + 1)-`"
        else
          keyset="`echo $keyset | cut -c $(expr $ref_index + 1)-`"
        fi
      done
      show "<tr><th>$key_index. $_httpdns_header_key</th><td>$key</td></tr>"
    done
  show "</table>"
}

show_PollingLinks()
{
  serverindex=$1
  # ------------------------
  if [ -n "$serverindex" ]
  then
    show "`show_tab_header \"$_httpdns_table_pollinghosts\" no`"
      show "<table class=\"normtable\">"
        show "<tr>"
          show "<th>$_httpdns_header_server</th>"
          show "<th>$_httpdns_header_status</th>"
        show "</tr>"
        show "<tr>"
          query_Server $serverindex "pollhosts"
        show "</tr>"
      show "</table>"
    show "`show_tab_footer`"
  fi
}

show_KillHost()
{
  serverindex=$1
  hostname=$2
  # ------------------------
  if [ -n "$serverindex" ]
  then
    if [ -n "$hostname" ]
    then
      show "`show_tab_header \"$_httpdns_table_killhost '$hostname'\" no`"
    else
      show "`show_tab_header \"$_httpdns_table_pollinghosts\" no`"
    fi
      show "<table class=\"normtable\">"
        show "<tr>"
          show "<th>$_httpdns_header_server</th>"
          show "<th>$_httpdns_header_status</th>"
        show "</tr>"
        show "<tr>"
          query_Server $serverindex "killhost" "$hostname"
        show "</tr>"
      show "</table>"
    show "`show_tab_footer`"
  fi
}

show_Sync()
{
  serverindex=$1 # (optional; queries all servers if empty)
  # ------------------------
  serverindexlist=""
  if [ -z "$serverindex" ]
  then
    serverindexlist="`seq 1 $HTTPDNS_SERVER_N`"
  else
    if [ 0$serverindex -le $HTTPDNS_SERVER_N ]
    then
      serverindexlist=$serverindex
    else
      show "`show_error \"$_httpdns_table_sync\" \"$_httpdns_error_badservernum ($serverindex)\"`"
    fi
  fi
  show "`show_tab_header \"$_httpdns_table_sync\" no`"
  show "<table class=\"normtable\">"
    show "<tr>"
      show "<th>$_httpdns_header_server</th>"
      show "<th>$_httpdns_header_status</th>"
    show "</tr>"
  for serverindex in $serverindexlist
  do
    show "<tr>"
      query_Server $serverindex
    show "</tr>"
  done
  show "</table>"
  show "`show_tab_footer`"
  show "<br>"
  update_HostsFile "yes"
}

update_HostsFile()
{
  verbose=$1
  # ------------------------
  sed -n '/^#/p' $HTTPDNS_HOSTSFILE > $HTTPDNS_TEMP
  cat $HTTPDNS_DNSFOLDER/*$HTTPDNS_DNSFILEEXTENSION | sort | sed '$!N; /^\([^ ]*\) .*\n\1 .*$/!P; D' | sed -n 's/^\([^ ]*\) \([0-9]*\) \([0-9.]*\)$/\3 \1/p' >> $HTTPDNS_TEMP
  if [ -n "$verbose" ]
  then
    show "`show_tab_header \"$_httpdns_table_hosts\" no`"
      show "<table class=\"normtable\">"
        show "<tr>"
          show "<th>$_httpdns_header_name</th>"
          show "<th>$_httpdns_header_ipaddress</th>"
          show "<th>$_httpdns_header_status</th>"
        show "</tr>"
        nomatch=""
        sed -n "/[0-9.] [^ ]*$/p" $HTTPDNS_TEMP | while read ip hostname
        do
          show "<tr>"
            show "<td>$hostname</td><td>$ip</td>"
            sed -n "/[0-9.] $hostname$/p" $HTTPDNS_HOSTSFILE | { read old_ip dummy
              if [ -n "$old_ip" ]
              then
                if [ "$ip" = "$old_ip" ]
                then
                  show "<td>$_httpdns_entry_unchanged</td>"
                else
                  show "<td class=\"httpdns_pass\">$_httpdns_entry_updated</td>"
                fi
              else
                show "<td class=\"httpdns_pass\">$_httpdns_entry_added</td>"
              fi
              }
          show "</tr>"
          [ 0 -eq 1 ]
        done && nomatch="yes"
        sed -n "/[0-9.] [^ ]*$/p" $HTTPDNS_HOSTSFILE | while read ip hostname
        do
          sed -n "/[0-9.] $hostname$/p" $HTTPDNS_TEMP | { read old_ip dummy
            if [ -z "$old_ip" ]
            then
              show "<tr><td>$hostname</td><td>$ip</td><td class=\"httpdns_fail\">$_httpdns_entry_deleted</td></tr>"
            fi
            }
          [ 0 -eq 1 ]
        done && if [ -n "$nomatch" ]; then show "<tr><td colspan=3><b>$_httpdns_status_nodata</b></td></tr>"; fi
      show "</table>"
    show "`show_tab_footer`"
  fi

  mv $HTTPDNS_TEMP $HTTPDNS_HOSTSFILE
  if [ -f $HTTPDNS_DNSMASQPID ]
  then
    dnsmasq_pid="`cat $HTTPDNS_DNSMASQPID`"
    if [ -n "$dnsmasq_pid" ]; then kill -SIGHUP $dnsmasq_pid; fi
  fi
  log " - Hosts file updated"
}

show_Journal()
{
  jlines=$1
  # ------------------------
  jlines=`echo $jlines | sed -n 's/^.[0-9]*$/&/p'`
  if [ 0$jlines -le 0 ]; then jlines=25; fi
  show "<h2><a href=\"$myname?download=journal\">$HTTPDNS_LOGFILE</a></h2>"
  show "<form action=\"$myname\" method=\"GET\">"
    show "<input type=\"hidden\" name=\"action\" value=\"$FORM_action\">"
    show "<input class=\"itext\" type=\"text\" size=\"5\" maxlength=\"5\" name=\"lines\" value=\"$jlines\">&nbsp;$_MN_rows&nbsp;&nbsp;"
    show "<input class=\"sbutton\" type=\"submit\" value=\"$_MN_show\">"
    show "<input class=\"sbutton\" type=\"submit\" name=\"showall\" value=\"$_MN_BTN_showallrows\">"
    show "<input class=\"sbutton\" type=\"submit\" name=\"reset\" value=\"$_MN_reset\">"
    show "<br>"
  show "</form>"
  show "<table class=\"normtable\">"
    show "<tr>"
      show "<th>$_httpdns_header_timestamp</th>"
      show "<th>$_httpdns_header_message</th>"
    show "</tr>"
    if [ -f "$HTTPDNS_LOGFILE" ]
    then
      tail -n $jlines $HTTPDNS_LOGFILE | while read line
      do
        show "`echo $line | sed -n 's|\(.[A-Za-z]* .[0-9]* .[0-9:]*\) \(.[^ ]*\) \(.[^ ]*:\) \(.*\)$|<tr><td>\1</td><td>\4</td></tr>|p'`"
        [ 0 -eq 1 ]
      done && show "<tr><td colspan=2><b>$_httpdns_status_nodata</b></td></tr>"
    else
      show "<tr><td colspan=2><b>$_httpdns_status_nodata</b></td></tr>"
    fi
  show "</table>"
}

query_Server() # MODIFIES: httpdnsheaderline
{
  serverindex=$1
  command=$2 # optional; syncs if empty
  hostname=$3 # optional
  # ------------------------
  httpdnsheaderline=""
  servername="`get_Servername $serverindex`"
  serverkey="`get_HTTPDNS_SERVER_ $serverindex _KEY`"
  localip="`get_LocalIP`"
  if [ -z "$localip" ]; then localip="`get_DummyIP`"; fi
  case "$command" in
    "" | "sync") if [ -n "$HTTPDNS_INTERFACE" ]; then magic="synl"; else magic="synk"; fi;;
    "version") magic="vrsn";;
    "viewlog") magic="vlog";;
    "clearlog") magic="klog";;
    "pollhosts") magic="list";;
    "killhost") if [ -n "$hostname" ]; then magic="kill"; else magic="list"; fi;;
    *) magic="";;
  esac
  case "$command" in
    "getlocalip")
      encrypted="no"
      query=""
    ;;
    *)
      if [ -n "$hostname" ]
      then
        encrypted="`get_Encrypted $localip \"$hostname\" \"$serverkey\" $magic`"
      else
        encrypted="`get_Encrypted $localip \"$HTTPDNS_HOSTNAME\" \"$serverkey\" $magic`"
      fi
      query="?query=$encrypted"
    ;;
  esac
  show "<td>$servername</td>"
  log "Quering $command on $servername ..."
  if [ -n "$encrypted" ]
  then
    if [ -z "`nslookup $servername | grep 'Unknown host'`" ]
    then
      case "$HTTPDNS_USEWGET" in
        "yes")
          query="`echo \"$servername\" | sed -n 's|^\(http[s]*://[^/]*\)\(.*\)|\1:'$HTTPDNS_QUERYPORT'\2|p'`/$HTTPDNS_SERVERFILE$query"
          options="-q -O $HTTPDNS_TEMP"
          # --- SSL variant
          # query="`echo \"$servername\" | sed -n 's|^\(http[s]*://[^/]*\)\(.*\)|\1\2|p'`/$HTTPDNS_SERVERFILE$query"
          # options="-q --no-check-certificate -O $HTTPDNS_TEMP"
          httpuser="`get_HTTPDNS_SERVER_ $serverindex _HTTPUSER`"
          if [ -n "$httpuser" ]; then options="$options --http-user=$httpuser"; fi
          httppassword="`get_HTTPDNS_SERVER_ $serverindex _HTTPPWD`"
          if [ -n "$httppassword" ]; then options="$options --http-password=$httppassword"; fi
          if [ "$HTTPDNS_DEBUG" = "yes" ]; then echo "$HTTPDNS_WGET $options \"$query\"" > $HTTPDNS_DEBUG_REQUEST; fi
          $HTTPDNS_WGET $options "$query"
        ;;
        *)
          serverdomain="`echo \"$servername\" | sed -n 's|^http[s]*://\([^/]*\).*|\1|p'`"
          echo "GET $servername/$HTTPDNS_SERVERFILE$query HTTP/1.0" > $HTTPDNS_TEMP
          echo "Host: $serverdomain" >> $HTTPDNS_TEMP
          echo "" >> $HTTPDNS_TEMP
          if [ "$HTTPDNS_DEBUG" = "yes" ]; then cp $HTTPDNS_TEMP $HTTPDNS_DEBUG_REQUEST; fi
          cat $HTTPDNS_TEMP | $HTTPDNS_NETCAT $serverdomain $HTTPDNS_QUERYPORT > $HTTPDNS_TEMP
        ;;
      esac
      sed -n 's/<[\/]*body>/\n&\n/;/<body>/,/<\/body>/p' $HTTPDNS_TEMP | tr "\n\15" "  " | sed -n 's/<br>/\n/g;s/<[^>]*>//gp;' | sed 's/^ *//;/^$/d' > $HTTPDNS_TEMP
      httpdnsheaderline="`grep $HTTPDNS_PACKAGE $HTTPDNS_TEMP`"
      if [ -n "$httpdnsheaderline" ]
      then
        if [ -n "`get_ServerVersion`" ]
        then
          response="`get_ServerResponse`"
          if [ -z "`echo $response | grep 'ERROR:'`" ]
          then
            if [ -z "`echo $response | grep 'ACCESS DENIED.'`" ]
            then
              query_Server_Parse "$command" "$servername" "$serverkey" "$hostname"
            else
              show "<td class=\"httpdns_fail\">$_httpdns_status_badkey</td>"
              log " - ERROR Bad serverkey"
              errorlevel 1
            fi
          else
            query_Server_Error "$response"
          fi
        else
          show "<td class=\"httpdns_fail\">$_httpdns_status_badresponse</td>"
          log " - ERROR Bad response"
          if [ "$HTTPDNS_DEBUG" = "yes" ]; then log " - Look into $HTTPDNS_DEBUG_RESPONSE for more info"; fi
          errorlevel 1
        fi
      else
        show "<td class=\"httpdns_fail\">$_httpdns_status_noresponse</td>"
        log " - ERROR Timeout or empty response"
        errorlevel 1
      fi
      if [ "$HTTPDNS_DEBUG" = "yes" ]; then cp $HTTPDNS_TEMP $HTTPDNS_DEBUG_RESPONSE; fi
      erase_file $HTTPDNS_TEMP
    else
      show "<td class=\"httpdns_fail\">$_httpdns_status_noaccess</td>"
      log " - ERROR Not reachable"
      errorlevel 1
    fi
  else
    show "<td class=\"httpdns_fail\">$_httpdns_status_encryption</td>"
    log " - ERROR Encryption of request failed"
    errorlevel 1
  fi
}

query_Server_Parse()
{
  command=$1
  servername=$2
  serverkey=$3
  hostname=$4
  # ------------------------
  case "$command" in
    "" | "sync")
      get_DecryptedFile $HTTPDNS_TEMP "$serverkey" > `get_FilenameFromServername "$servername"`
      show "<td class=\"httpdns_pass\">$_httpdns_status_synced</td>"
      log " - Synced successfully"
    ;;
    "version")
      show "<td>`get_ServerType` `get_ServerVersion`</td>"
      log " - Received version info `get_ServerType` `get_ServerVersion`"
    ;;
    "viewlog")
            show "<td class=\"httpdns_pass\">$_httpdns_status_success</td>"
          show "</tr>"
        show "</table>"
      show "`show_tab_footer`"
      show "<br>"
      show "`show_tab_header \"$servername\" no`"
        show "<table class=\"normtable\">"
          show "<tr>"
            show "<th>$_httpdns_header_datetime</th>"
            show "<th>$_httpdns_header_name</th>"
            show "<th>$_httpdns_header_message</th>"
          show "</tr>"
          show "<tr>"
        found=""
        sed -n '2,$ p' $HTTPDNS_TEMP | while read line
        do
          row="`echo $line | sed -n 's/^\([0-9-]*\) \([0-9:]*\) \([^ ]*\) \(.*\)/<tr><td>\1 \2<\/td><td>\3<\/td><td>\4<\/td><\/tr>/p'`"
          if [ -n "$row" ]
          then
            show "$row"
            [ 0 -eq 1 ]
          fi
        done && show "<tr><td colspan=3><b>$_httpdns_status_nodata</b></td></tr>"
      log " - Received server log"
    ;;
    "clearlog")
      show "<td class=\"httpdns_pass\">$_httpdns_status_success</td>"
      log " - Cleared server log"
    ;;
    "pollhosts")
            show "<td class=\"httpdns_pass\">$_httpdns_status_success</td>"
          show "</tr>"
        show "</table>"
      show "`show_tab_footer`"
      show "<br>"
      show "`show_tab_header \"$servername\" no`"
        show "<table class=\"normtable\">"
          show "<tr>"
            show "<th>$_httpdns_header_name</th>"
            show "<th>$_httpdns_header_link</th>"
          show "</tr>"
        get_DecryptedFile $HTTPDNS_TEMP "$serverkey" | cut -d' ' -f1 | while read host
        do
          hostlink="$servername/$HTTPDNS_SERVERFILE?query=`get_Encrypted $(get_DummyIP) \"$host\" \"$serverkey\" poll`"
          show "<tr><td>$host</td><td><a href=\"$hostlink\">$hostlink</a></td></tr>"
          [ 0 -eq 1 ]
        done && show "<tr><td colspan=2><b>$_httpdns_status_nodata</b></td></tr>"
      log " - Received host info"
    ;;
    "killhost")
      if [ -n "$hostname" ]
      then
        show "<td class=\"httpdns_pass\">$_httpdns_status_success</td>"
        log " - DNS entry '$hostname' erased."
      else
              show "<td class=\"httpdns_pass\">$_httpdns_status_success</td>"
            show "</tr>"
          show "</table>"
        show "`show_tab_footer`"
        show "<br>"
        show "`show_tab_header \"$servername\" no`"
          show "<table class=\"normtable\">"
            show "<tr>"
              show "<th>$_httpdns_header_name</th>"
              show "<th>$_httpdns_header_ipaddress</th>"
              show "<th>$_httpdns_header_timestamp</th>"
              show "<th>$_httpdns_header_command</th>"
            show "</tr>"
          timezone="`date +%z | sed -n 's/[+-]/ & /p'`"
          get_DecryptedFile $HTTPDNS_TEMP "$serverkey" | while read hostname timestamp ipaddress
          do
            show "<tr>"
              show "<td>$hostname</td><td>$ipaddress</td><td>`get_LocalizedTime \"$timezone\" \"$timestamp\"`</td>"
              show "<td><a href=\"$myname?action=killhost&server=$serverindex&hostname=$hostname\"><img src=\"../img/httpdns_delete.gif\" border=0 title=\"$_httpdns_button_killhost\"></a></td>"
            show "</tr>"
            [ 0 -eq 1 ]
          done && show "<tr><td colspan=4><b>$_httpdns_status_nodata</b></td></tr>"
        log " - Received host info"
      fi
    ;;
    "getlocalip")
      show "<td>`get_ServerResponse`</td>"
      log " - Received WAN ip address `get_ServerResponse`"
    ;;
  esac
}

query_Server_Error()
{
  response=$1
  # ------------------------
  case "`echo $response | sed -n 's/.*ERROR: \(.*\)$/\1/p'`" in
    "Logging not activated.")
      show "<td class=\"httpdns_fail\">$_httpdns_status_logingdisabled</td>"
      log " - ERROR logging not activated on server"
      errorlevel 1
    ;;
    "Cannot rewrite Log.")
      show "<td class=\"httpdns_fail\">$_httpdns_status_logwritefailed</td>"
      log " - ERROR Cannot write to log file or database on server"
      errorlevel 1
    ;;
    "Cannot open DNS file or table.")
      show "<td class=\"httpdns_fail\">$_httpdns_status_dnsfilemissing</td>"
      log " - ERROR Cannot read DNS file or database on server"
      errorlevel 1
    ;;
    "Unrecognized command.")
      show "<td class=\"httpdns_fail\">$_httpdns_status_badcommand</td>"
      log " - ERROR Server did not understand the command"
      errorlevel 1
    ;;
    "No access to DNS or log folder.")
      show "<td class=\"httpdns_fail\">$_httpdns_status_foldersmissing</td>"
      log " - ERROR DNS or log folder inaccessible on server"
      errorlevel 1
    ;;
    "Cannot read log.")
      show "<td class=\"httpdns_fail\">$_httpdns_status_cannotreadlog</td>"
      log " - ERROR Cannot read log file or database on server"
      errorlevel 1
    ;;
    "MySQL login failed.")
      show "<td class=\"httpdns_fail\">$_httpdns_status_mysqlloginfailed</td>"
      log " - ERROR Login to MySQL database failed"
      errorlevel 1
    ;;
    "Cannot erase DNS entry.")
      show "<td class=\"httpdns_fail\">$_httpdns_status_cannoterasednsentry</td>"
      log " - ERROR Cannot erase DNS entry"
      errorlevel 1
    ;;
    "Unauthorized Administator Command.")
      show "<td class=\"httpdns_fail\">$_httpdns_status_unauthorizedadmincmd</td>"
      log " - ERROR Unauthorized Administator Command."
      errorlevel 1
    ;;
    *)
      show "<td class=\"httpdns_fail\">$_httpdns_status_servererror</td>"
      log " - ERROR General failure"
      errorlevel 1
    ;;
  esac
}

upload_ServerProgram()
{
  serverindex=$1
  # ------------------------
  log "Server Programm FTP upload"
  if [ -n "$serverindex" -a 0$serverindex -le $HTTPDNS_SERVER_N ]
  then
    if [ -f $HTTPDNS_SERVER ]
    then
      if [ -f $HTTPDNS_FTPPROGRAM ]
      then
        ftp_url="`get_HTTPDNS_SERVER_ $serverindex _FTPURL | sed 's/\/$//'`"
        ftp_user="`get_HTTPDNS_SERVER_ $serverindex _FTPUSER`"
        ftp_pwd="`get_HTTPDNS_SERVER_ $serverindex _FTPPWD`"
        if [ -n "$ftp_url" -a -n "$ftp_user" -a -n "$ftp_pwd" ]
        then
          log " - Patching server program"
          patch_ServerProgram $serverindex
          if [ -f $HTTPDNS_SERVERTEMP ]
          then
            show "`show_tab_header \"$_httpdns_table_upload\" no`"
            show "<table class=\"normtable\">"
              show "<tr>"
                show "<th>$_httpdns_header_server</th>"
                show "<th>$_httpdns_header_status</th>"
              show "</tr>"
              show "<tr>"
                show "<td>$ftp_url</td>"
                if [ -z "`nslookup $ftp_url | grep 'Unknown host'`" ]
                then
                  logfile="`get_HTTPDNS_SERVER_ $serverindex _LOGFILE`"
                  dnsfile="`get_HTTPDNS_SERVER_ $serverindex _DNSFILE`"
                  if [ -z "$dnsfile" ]; then dnsfile=$HTTPDNS_SERVER_DEFAULT_DNSFILE; fi
                  log " - Creating FTP command list"
                  get_FTPCommandList $serverindex "$ftp_url" "$ftp_user" "$ftp_pwd" "$dnsfile" "$logfile"> $HTTPDNS_TEMP
                  log " - Uploading to $ftp_url"
                  cat $HTTPDNS_TEMP | $HTTPDNS_FTPPROGRAM -vn > $HTTPDNS_TEMP
                  if [ -n "`grep '^230' $HTTPDNS_TEMP`" ]
                  then
                    if [ -n "`grep '^226' $HTTPDNS_TEMP`" ]
                    then
                      show "<td class=\"httpdns_pass\">$_httpdns_status_success</td>"
                      log " - Finished successfully"
                    else
                      show "<td class=\"httpdns_fail\">$_httpdns_status_ftptransferfailed</td>"
                      log " - ERROR File transfer failed"
                      errorlevel 1
                    fi
                  else
                    show "<td class=\"httpdns_fail\">$_httpdns_status_ftploginfailed</td>"
                    log " - ERROR FTP login failed"
                    errorlevel 1
                  fi
                    show "</tr>"
                  show "</table>"
                  show "<table>"
                    show "<td>"
                    show "`sed 's,^[4-5][0-9][0-9] .*$,<b>&</b>,' $HTTPDNS_TEMP | sed -n 's/^.*$/&<br>/p'`"
                    show "</td>"
                  show "</table>"
                  erase_file $HTTPDNS_TEMP
                  erase_file "$HTTPDNS_TEMP_FOLDER/$(get_FileName $logfile)"
                  erase_file "$HTTPDNS_TEMP_FOLDER/$(get_FileName $dnsfile)"
                else
                  show "<td class=\"httpdns_fail\">$_httpdns_status_noaccess</td>"
                  log " - ERROR $ftp_url not reachable"
                  errorlevel 1
                fi
              show "</tr>"
            show "</table>"
            show "`show_tab_footer`"
            show "<br>"
            erase_file $HTTPDNS_SERVERTEMP
          else
            show "`show_error \"$_httpdns_table_upload\" \"$_httpdns_error_nokey\"`"
            log " - ERROR Empty server key"
            errorlevel 1
          fi
        else
          show "`show_error \"$_httpdns_table_upload\" \"$_httpdns_error_noftppassword\"`"
          log " - ERROR FTP URL, username or password is empty"
          errorlevel 1
        fi
      else
        show "`show_error \"$_httpdns_table_upload\" \"$_httpdns_error_ftpfilenotfound\"`"
        log " - ERROR FTP program is missing"
        errorlevel 1
      fi
    else
      show "`show_error \"$_httpdns_table_upload\" \"$_httpdns_error_cgifilenotfound\"`"
      log " - ERROR File $HTTPDNS_SERVER is missing"
      errorlevel 1
    fi
  else
    show "`show_error \"$_httpdns_table_upload\" \"$_httpdns_error_badservernum ($serverindex)\"`"
    log " - ERROR Illegal server number"
    errorlevel 1
  fi
}

get_FTPCommandList() # OUTPUT: ftp command file
{
  serverindex=$1
  ftp_url=$2
  ftp_user=$3
  ftp_pwd=$4
  dnsfile=$5
  logfile=$6
  # ------------------------
  ftp_folder="`get_HTTPDNS_SERVER_ $serverindex _FOLDER | sed 's/\/$//'`"
  case "`get_HTTPDNS_SERVER_ $serverindex _MYSQL`" in
    "yes")
      mysql_dnstable="`get_HTTPDNS_SERVER_ $serverindex _MYSQL_DNSTABLE`"
      mysql_logtable="`get_HTTPDNS_SERVER_ $serverindex _MYSQL_LOGTABLE`"
    ;;
    *)
      mysql_dnstable=""
      mysql_logtable=""
    ;;
  esac
  echo "open $ftp_url"
  echo "user $ftp_user $ftp_pwd"
  while [ -n "$ftp_folder" ]
  do
    folder="`echo $ftp_folder | cut -d'/' -f1`"
    echo "mkdir $folder"
    echo "chmod 755 $folder"
    echo "cd $folder"
    ftp_folder="`echo $ftp_folder | sed -n 's/[^/]*\/\(.*\)/\1/p'`"
  done
  echo "put $HTTPDNS_SERVERTEMP $HTTPDNS_SERVERFILE"
  echo "chmod 644 $HTTPDNS_SERVERFILE"
  if [ -z "$mysql_logtable" ]; then get_FTPCommandList_CreateFolder "$logfile"; fi
  if [ -z "$mysql_dnstable" ]; then get_FTPCommandList_CreateFolder "$dnsfile"; fi
  echo "bye"
}

get_FTPCommandList_CreateFolder()
{
  pathandfilename=$1
  # ------------------------
  if [ -n "$pathandfilename" ]
  then
    folder=""
    for subfolder in `get_Path $pathandfilename | tr '/' ' '`
    do
      folder="$folder$subfolder"
      if [ -n "$folder" ]
      then
        echo "mkdir $folder"
        echo "chmod 755 $folder"
        folder="$folder/"
      fi
    done
    filename="`get_FileName $pathandfilename`"
    > $HTTPDNS_TEMP_FOLDER/$filename
    echo "put $HTTPDNS_TEMP_FOLDER/$filename $folder$filename"
    echo "chmod 777 $folder$filename"
  fi
}

patch_ServerProgram()
{
  serverindex=$1
  # ------------------------
  config_dnsfilename="`get_HTTPDNS_SERVER_ $serverindex _DNSFILE`"
  if [ -z "$config_dnsfilename" ]; then config_dnsfilename=$HTTPDNS_SERVER_DEFAULT_DNSFILE; fi
  config_logfilename="`get_HTTPDNS_SERVER_ $serverindex _LOGFILE`"
  config_serverkey="`get_HTTPDNS_SERVER_ $serverindex _KEY`"
  config_serverpass="`get_HTTPDNS_SERVER_ $serverindex _PASSWORD`"
  config_loglevel=`get_HTTPDNS_SERVER_ $serverindex _LOGLEVEL`
  if [ -z "$config_loglevel" ]; then config_loglevel=$HTTPDNS_SERVER_DEFAULT_LOGLEVEL; fi
  config_lifetime=`get_HTTPDNS_SERVER_ $serverindex _LIFETIME`
  if [ -z "$config_lifetime" ]; then config_lifetime=$HTTPDNS_SERVER_DEFAULT_LIFETIME; fi
  config_admin="`get_HTTPDNS_SERVER_ $serverindex _ADMIN`"
  config_mysql="`get_HTTPDNS_SERVER_ $serverindex _MYSQL`"
  case "$config_mysql" in
    "yes")
      config_mysql_server="`get_HTTPDNS_SERVER_ $serverindex _MYSQL_SERVER`"
      config_mysql_username="`get_HTTPDNS_SERVER_ $serverindex _MYSQL_USER`"
      config_mysql_password="`get_HTTPDNS_SERVER_ $serverindex _MYSQL_PWD`"
      config_mysql_database="`get_HTTPDNS_SERVER_ $serverindex _MYSQL_DB`"
      config_mysql_dnstable="`get_HTTPDNS_SERVER_ $serverindex _MYSQL_DNSTABLE`"
      config_mysql_logtable="`get_HTTPDNS_SERVER_ $serverindex _MYSQL_LOGTABLE`"
      if [ -n "$config_mysql_dnstable" ]
      then
        config_dnsfilename=""
        log "   - Using MySQL table $config_mysql_dnstable for DNS"
      else
        log "   - Using file $config_dnsfilename for DNS"
      fi
      if [ -n "$config_mysql_logtable" ]
      then
        config_logfilename=""
        log "   - Using MySQL table $config_mysql_logtable for logging"
      else
        log "   - Using file $config_logfilename for logging"
      fi
    ;;
    *)
    config_mysql_server=""
    config_mysql_username=""
    config_mysql_password=""
    config_mysql_database=""
    config_mysql_dnstable=""
    config_mysql_logtable=""
    ;;
  esac
  config_serverkey="`get_HTTPDNS_SERVER_ $serverindex _KEY | sed -n \"s/./,'&'/gp\" | cut -c2-`"
  if [ -n "$config_serverkey" ]
  then
    cat >$HTTPDNS_TEMP <<EOF
    s;//config_logfilename//;$config_logfilename;
    s;//config_dnsfilename//;$config_dnsfilename;
    s;//config_serverkey//;$config_serverkey;
    s;//config_loglevel//;$config_loglevel;
    s;//config_lifetime//;$config_lifetime;
    s;//config_admin//;$config_admin;
    s;//config_serverpass//;$config_serverpass;
    s;//config_mysql_server//;$config_mysql_server;
    s;//config_mysql_username//;$config_mysql_username;
    s;//config_mysql_password//;$config_mysql_password;
    s;//config_mysql_database//;$config_mysql_database;
    s;//config_mysql_dnstable//;$config_mysql_dnstable;
    s;//config_mysql_logtable//;$config_mysql_logtable;
EOF
    sed -f $HTTPDNS_TEMP <$HTTPDNS_SERVER >$HTTPDNS_SERVERTEMP
    erase_file $HTTPDNS_TEMP
  fi
}

do_Download()
{
  file=$1
  # ------------------------
  case "$file" in
    "journal") source="$HTTPDNS_LOGFILE";;
    "server") source="$HTTPDNS_SERVERTEMP";;
    *) source="";;
  esac
  if [ -n "$source" -a -f $source ]
  then
    http_header download "ctype=text/plain;filename=`echo \"$source\" | sed -n 's|.*/\(.*\)|\1|p'`"
    cat $source
  else
    show_html_header "$HTTPDNS_PACKAGE $HTTPDNS_VERSION"
    show_error "$HTTPDNS_PACKAGE" "$_httpdns_error_downloadfilenotfound ($source)"
    show_html_footer
  fi
}

do_Action()
{
  action=$1
  serverindex=$2 # optional
  hostname=$3    # optional
  # ------------------------
  case "$action" in
    "sync" | "keygen")
      check_rights "$HTTPDNS_RIGHTS_SECTION" "view"
    ;;
    "query")
      check_rights "$HTTPDNS_RIGHTS_SECTION" "sync"
    ;;
    "upload" | "manual" | "admin" | "viewlog" | "clearlog" | "journal")
      check_rights "$HTTPDNS_RIGHTS_SECTION" "admin"
    ;;
    *)
    ;;
  esac

  show_html_header "$HTTPDNS_PACKAGE $HTTPDNS_VERSION"
  _sync=$_httpdns_tab_sync; _keygen=$_httpdns_tab_keygen; _admin=$_httpdns_tab_admin; _journal=$_httpdns_tab_journal
  if [ -n "`check_UserRight admin`" ]
  then
    if [ -n "$HTTPDNS_LOGFILE" ]
    then
      show_tab_header "$_sync" "$myname?action=sync" "$_keygen" "$myname?action=keygen" "$_admin" "$myname?action=admin" "$_journal" "$myname?action=journal"
    else
      show_tab_header "$_sync" "$myname?action=sync" "$_keygen" "$myname?action=keygen" "$_admin" "$myname?action=admin"
    fi
  else
    show_tab_header "$_sync" "$myname?action=sync" "$_keygen" "$myname?action=keygen"
  fi

  case "$action" in
    "sync")
      show_LocalIP
      show_DNSTables
    ;;
    "query")
      show_Sync $serverindex
    ;;
    "upload")
      upload_ServerProgram $serverindex
    ;;
    "manual")
      log "Patching server program for manual upload"
      patch_ServerProgram $serverindex
      if [ -f $HTTPDNS_SERVERTEMP ]
      then
        show_info "$_httpdns_table_manual" "$_httpdns_info_manualinstall"
      else
        show_error "$_httpdns_table_manual" "$_httpdns_error_nokey"
      fi
    ;;
    "keygen")
      show_RandomKeys
    ;;
    "linkgen")
      show_PollingLinks $serverindex
    ;;
    "killhost")
      show_KillHost $serverindex "$hostname"
    ;;
    "admin")
      show_AdminTable
    ;;
    "viewlog")
      show_ServerLog $serverindex "viewlog"
    ;;
    "clearlog")
      show_ServerLog $serverindex "clearlog"
    ;;
    "journal")
      if [ -n "$FORM_reset" -a -f $HTTPDNS_LOGFILE ]; then rm $HTTPDNS_LOGFILE; fi
      if [ -n "$FORM_showall" ]
      then
          show_Journal 65535
      else
          show_Journal $FORM_lines
      fi
    ;;
    *)
      show_error "$HTTPDNS_PACKAGE" "$_httpdns_error_badaction ($action)"
    ;;
  esac
  show_tab_footer
  show_html_footer
}

do_Command()
{
  # uses any number of parameters
  # ------------------------
  updatehostsfile=""
  serverindexlist=""
  unknownserver=""
  matchingserver=""
  for option in "$@"
  do
    case "$option" in
      "--help" | "-h")
        echo "$HTTPDNS_PACKAGE $HTTPDNS_VERSION"
        echo ""
        echo "usage: httpdns.cgi -h"
        echo "       httpdns.cgi -a|-i|-u [-s]"
        echo "       httpdns.cgi [-s] {server}"
        echo ""
        echo "  -a, --all     query all servers"
        echo "  -h, --help    display this message (cancels all other commands)"
        echo "  -i, --ipup    query servers with _IPUP set to 'yes'"
        echo "  -u, --update  update hosts file, only (no server querying)"
        echo "  -s, --silent  do not output to console"
        echo "  {server}      query one particular server given by its URL or"
        echo "                its server number as with HTTPDNS_SERVER_N."
        echo ""
        echo "The router has to be online and at least one server hast to be"
        echo "configured for the program to work."
        serverindexlist=""
        matchingserver="yes"
        httpdnsverbose=""
        break
      ;;
      "--update" | "-u")
        updatehostsfile="yes"
        matchingserver="yes"
      ;;
      "--silent" | "-s")
        httpdnsverbose="no"
      ;;
      "--all" | "-a")
        serverindexlist="`seq 1 $HTTPDNS_SERVER_N`"
        matchingserver="yes"
      ;;
      "--ipup" | "-i")
        for serverindex in `seq 1 $HTTPDNS_SERVER_N`
        do
          if [ "`get_HTTPDNS_SERVER_ $serverindex _IPUP`" = "yes" ]
          then
            if [ -z "`echo \"$serverindexlist\" | grep \"$serverindex\"`" ]
            then
              serverindexlist="`echo -e \"$serverindexlist\n$serverindex\"`"
              matchingserver="yes"
            fi
          fi
        done
      ;;
      [1-9]|[1-9][0-9]|[1-9][0-9][0-9])
        if [ $option -le $HTTPDNS_SERVER_N ]
        then
          if [ -z "`echo \"$serverindexlist\" | grep \"$option\"`" ]
          then
            serverindexlist="`echo -e \"$serverindexlist\n$option\"`"
            matchingserver="yes"
          fi
        else
          unknownserver="`echo -e \"$unknownserver\n$option\"`"
        fi
      ;;
      *)
        match=""
        for serverindex in `seq 1 $HTTPDNS_SERVER_N`
        do
          if [ "`get_Servername $serverindex`" = "`echo $option | sed 's|/$||'`" ]
          then
            match="yes"
            if [ -z "`echo \"$serverindexlist\" | grep \"$serverindex\"`" ]
            then
              serverindexlist="`echo -e \"$serverindexlist\n$serverindex\"`"
              matchingserver="yes"
            fi
            break
          fi
        done
        if [ -z "$match" ]; then unknownserver="`echo -e \"$unknownserver\n$option\"`"; fi
      ;;
    esac
  done
  if [ "$httpdnsverbose" = "yes" ]; then echo "$HTTPDNS_PACKAGE $HTTPDNS_VERSION"; fi
  for serverindex in $serverindexlist
  do
    query_Server $serverindex "sync"
    updatehostsfile="yes"
  done
  if [ -n "$unknownserver" ]
  then
    for server in $unknownserver
    do
      log "ERROR: Unknown server '$server'."
      errorlevel 1
    done
  fi
  if [ -z "$matchingserver" ]
  then
    log "ERROR: No matching server."
    errorlevel 1
  fi
  if [ "$updatehostsfile" = "yes" ]
  then
    update_HostsFile
  fi
}

# ----------------------------------------------------------------------------

if [ -n "$FORM_download" ]
then
  do_Download "$FORM_download"
else
  if [ -n "$FORM_action" ]
  then
    do_Action "$FORM_action" $FORM_server "$FORM_hostname"
  else
    httpdnsverbose="yes"
    if [ $HTTPDNS_SERVER_N -gt 0 -a -n "$*" ]
    then
      do_Command "$@"
    else
      do_Command "--help"
    fi
  fi
fi
return $httpdnserrorlevel
