#Region ;**** Directives created by AutoIt3Wrapper_GUI ****
#AutoIt3Wrapper_icon=httpDNS-WinClient.ico
#AutoIt3Wrapper_outfile=..\httpDNS-WinClient.exe
#EndRegion ;**** Directives created by AutoIt3Wrapper_GUI ****
Opt("MustDeclareVars", 1)

;------------------------------------------------------------------------------
; httpDNS - Windows Client Program                                       1.0.0
;   Copyright (C) 2007-2008 Tiger
;
;   This OPT package is licensed under Free Software Foundation, Inc.,
;   General Public License Version 2, June 1991
;------------------------------------------------------------------------------

#include <INet.au3>
#include <GUIConstants.au3>
#include <File.au3>

Global $version = "httpDNS Windows Client 1.0.0" 

Const $GUIFont = "MS Sans Serif" 
Const $GUIFontSize = Default ; 8.5
Const $GUIFontWeight = 400
Const $GUILabelFont = $GUIFont
Const $GUILabelFontSize = $GUIFontSize
Const $GUILabelFontWeight = $GUIFontWeight * 2

Const $RegistryRoot = "HKEY_CURRENT_USER\SOFTWARE\httpDNS" 
Const $RegistrySettings = $RegistryRoot & "\Settings" 

Const $TempFilename = EnvGet("TEMP") & "\httpdns.tmp" 

Const $PatternIPAddress = "[[:digit:]]+.[[:digit:]]+.[[:digit:]]+.[[:digit:]]+" 
Const $PatternIP = "@IP" 
Const $PatternTime = "@TIME" 
Const $PatternDate = "@DATE" 

Global $Settings[10][3]
Const $SettingName = 0
Const $SettingValue = 1
Const $SettingHandle = 2
$Settings[0][$SettingName] = "URL_Query" 
$Settings[1][$SettingName] = "URL_Sync" 
$Settings[2][$SettingName] = "Hostname" 
$Settings[3][$SettingName] = "Password" 
$Settings[4][$SettingName] = "Filename" 
$Settings[5][$SettingName] = "Find" 
$Settings[6][$SettingName] = "ReplaceWith" 
$Settings[7][$SettingName] = "Append" 
$Settings[8][$SettingName] = "Execute" 
$Settings[9][$SettingName] = "Tab" 

#include <httpDNS-WinClient.en>

Func InitLanguage($os_language)
	Dim $result
	$result = False
	Switch $os_language
		Case "0407", "0807", "0c07", "1007", "1407" 
			#include <httpDNS-WinClient.de>
			$result = True
		Case "040c", "080c", "0c0c", "100c", "140c", "180c" 
			#include <httpDNS-WinClient.fr>
			$result = True
		Case Else
			;#include <httpDNS-WinClient.en>
			$result = True
	EndSwitch
	Return $result
EndFunc

Func ErrorMessage($message)
	MsgBox(16, $version, $message)
EndFunc

; ---------------------
; Handling of Settings

Func Setting($tag, $value = Default)
	Dim $index, $result
	$result = ""
	$index = 0
	While $index < UBound($Settings) And $Settings[$index][$SettingName] <> $tag
		$index = $index + 1
	WEnd
	If $Settings[$index][$SettingName] = $tag Then
		If $value <> Default Then
			$Settings[$index][$SettingValue] = $value
		EndIf
		$result = $Settings[$index][$SettingValue]
	Else
		ErrorMessage($_error_unknownsetting & " '" & $tag & "'.")
	EndIf
	Return $result
EndFunc

Func Handle($tag, $handle = Default)
	Dim $index, $result
	$result = ""
	$index = 0
	While $index < UBound($Settings) And $Settings[$index][$SettingName] <> $tag
		$index = $index + 1
	WEnd
	If $Settings[$index][$SettingName] = $tag Then
		If $handle <> Default Then
			$Settings[$index][$SettingHandle] = $handle
		EndIf
		$result = $Settings[$index][$SettingHandle]
	Else
		ErrorMessage($_error_unknownsetting & " '" & $tag & "'.")
	EndIf
	Return $result
EndFunc

Func SaveSettings($profile = Default)
	Dim $index
	If $profile = "" Or $profile = Default Or $profile = $_profile_new Then
		$profile = "Default" 
	EndIf
	RegWrite($RegistrySettings, "Default", "REG_SZ", $profile)
	For $index = 0 To UBound($Settings) - 1
		RegWrite($RegistrySettings & "\" & $profile, $Settings[$index][$SettingName], "REG_SZ", $Settings[$index][$SettingValue])
		If @error Then
			ErrorMessage($_error_writingsetting_1 & " '" & @error & "' " & $_error_writingsetting_2 & " '" & $Settings[$index][$SettingName] & "' " & $_error_writingsetting_3)
		EndIf
	Next
EndFunc

Func RestoreSettings($profile = Default)
	Dim $index
	If $profile = "" Or $profile = Default Then
		$profile = ReadProfileDefault()
		If $profile = "" Then
			$profile = "Default" 
		EndIf
	EndIf
	For $index = 0 To UBound($Settings) - 1
		$Settings[$index][$SettingValue] = RegRead($RegistrySettings & "\" & $profile, $Settings[$index][$SettingName])
		If @error Then
			$Settings[$index][$SettingValue] = ""
		EndIf
	Next
EndFunc

Func UpdateSettings()
	Dim $index
	For $index = 0 To UBound($Settings) - 1
		If $Settings[$index][$SettingHandle] <> 0 Then
			$Settings[$index][$SettingValue] = GUICtrlRead($Settings[$index][$SettingHandle])
		Else
			$Settings[$index][$SettingValue] = ""
		EndIf
	Next
EndFunc

; -----------------
; Profile Handling

Func ReadProfileDefault()
	Dim $result
	$result = RegRead($RegistrySettings, "Default")
	Return $result
EndFunc

Func ReadProfiles()
	Dim $profile, $profiles, $index
	$profiles = $_profile_new
	$index = 1
	Do
		$profile = RegEnumKey($RegistrySettings, $index)
		If Not @error Then
			$profiles = $profile & Opt("GUIDataSeparatorChar") & $profiles
			$index = $index + 1
		EndIf
	Until @error
	Return $profiles
EndFunc

; ----------------
; Dialog Routines

Func DialogButton($label, $x, $y, $width, $height, $tooltip = Default, $style = Default)
	Dim $result
	$result = GUICtrlCreateButton($label, $x, $y, $width, $height, $style)
	GUICtrlSetFont($result, $GUILabelFontSize, $GUILabelFontWeight, 0, $GUILabelFont)
	If $result <> 0 And $tooltip <> Default Then
		GUICtrlSetTip($result, $tooltip)
	EndIf
	Return $result
EndFunc

Func DialogButtonWithIcon($label, $x, $y, $width, $height, $tooltip = Default)
	Dim $result
	$result = DialogButton($label, $x, $y, $width, $height, $tooltip, $BS_ICON)
	GUICtrlSetImage($result, "shell32.dll", 5, 0)
	Return $result
EndFunc

Func DialogInput($data, $x, $y, $width, $height, $label, $label_x, $label_y, $label_width, $label_height, $tooltip = Default, $style = Default)
	Dim $result
	If GUICtrlCreateLabel($label, $label_x, $label_y, $label_width, $label_height) <> 0 Then
		GUICtrlSetFont(-1, $GUILabelFontSize, $GUILabelFontWeight, 0, $GUILabelFont)
		If $tooltip <> Default Then
			GUICtrlSetTip(-1, $tooltip)
		EndIf
	EndIf
	$result = GUICtrlCreateInput($data, $x, $y, $width, $height, $style)
	GUICtrlSetFont($result, $GUIFontSize, $GUIFontWeight, 0, $GUIFont)
	If $result <> 0 And $tooltip <> Default Then
		GUICtrlSetTip($result, $tooltip)
	EndIf
	Return $result
EndFunc

Func DialogCombo($data, $x, $y, $width, $height, $label, $label_x, $label_y, $label_width, $label_height, $tooltip = Default)
	Dim $result
	If GUICtrlCreateLabel($label, $label_x, $label_y, $label_width, $label_height) <> 0 Then
		GUICtrlSetFont(-1, $GUILabelFontSize, $GUILabelFontWeight, 0, $GUILabelFont)
		If $tooltip <> Default Then
			GUICtrlSetTip(-1, $tooltip)
		EndIf
	EndIf
	$result = GUICtrlCreateCombo($data, $x, $y, $width, $height)
	GUICtrlSetFont($result, $GUIFontSize, $GUIFontWeight, 0, $GUIFont)
	If $result <> 0 And $tooltip <> Default Then
		GUICtrlSetTip($result, $tooltip)
	EndIf
	Return $result
EndFunc

Func DialogCheckbox($label, $x, $y, $width, $height, $tooltip = Default)
	Dim $result
	$result = GUICtrlCreateCheckbox($label, $x, $y, $width, $height)
	GUICtrlSetFont($result, $GUILabelFontSize, $GUILabelFontWeight, 0, $GUILabelFont)
	If $result <> 0 And $tooltip <> Default Then
		GUICtrlSetTip($result, $tooltip)
	EndIf
	Return $result
EndFunc

Func UpdateDialog($Form_Profile, $Form_Query, $Form_Synch)
	Dim $profile
	$profile = GUICtrlRead($Form_Profile)
	If $profile = $_profile_new Then
		$profile = ""
	EndIf
	GUICtrlSetData($Form_Profile, "")
	GUICtrlSetData($Form_Profile, ReadProfiles(), $profile)
	GUICtrlSetData(Handle("URL_Sync"), Setting("URL_Sync"))
	GUICtrlSetData(Handle("Hostname"), Setting("Hostname"))
	GUICtrlSetData(Handle("Password"), Setting("Password"))
	GUICtrlSetData(Handle("URL_Query"), Setting("URL_Query"))
	GUICtrlSetData(Handle("Filename"), Setting("Filename"))
	GUICtrlSetData(Handle("Find"), Setting("Find"))
	GUICtrlSetData(Handle("ReplaceWith"), Setting("ReplaceWith"))
	GUICtrlSetData(Handle("Execute"), Setting("Execute"))
	GUICtrlSetState(Handle("Append"), Number(Setting("Append")))
	Switch Setting("Tab")
		Case "0" 
			GUICtrlSetState($Form_Synch, $GUI_SHOW)
		Case "1" 
			GUICtrlSetState($Form_Query, $GUI_SHOW)
	EndSwitch
EndFunc

Func Dialog($profile)
	Dim $Form, $Form_Profile, $Form_Synch, $Form_Query
	Dim $Form_Abort, $Form_Submit, $Form_SaveProfile, $Form_SelectFilename, $Form_SelectExecute
	$Form = GUICreate($version, 626, 306, -1, -1)
	$Form_Submit = DialogButton($_form_submit, 338, 265, 75, 25, $_form_submit_tt)
	$Form_Abort = DialogButton($_form_abort, 212, 265, 75, 25, $_form_abort_tt)
	$Form_Profile = DialogCombo($profile, 114, 5, 430, 200, $_form_profile, 12, 8, 100, 20, $_form_profile_tt)
	$Form_SaveProfile = DialogButton($_form_save, 554, 5, 60, 25, $_form_save_tt)
	Handle("Tab", GUICtrlCreateTab(0, 37, 625, 223))
	GUICtrlSetFont(Handle("Tab"), $GUILabelFontSize, $GUILabelFontWeight, 0, $GUILabelFont)
	GUICtrlSetTip(Handle("Tab"), $_form_selector_tt)
	$Form_Synch = GUICtrlCreateTabItem($_form_tab_synchronize)
	Handle("URL_Sync", DialogInput("", 114, 72, 500, 25, $_form_serverurl, 12, 75, 100, 20, $_form_serverurl_tt_sync))
	Handle("Hostname", DialogInput("", 114, 107, 250, 24, $_form_hostname, 12, 110, 84, 20, $_form_hostname_tt))
	Handle("Password", DialogInput("", 114, 132, 250, 24, $_form_password, 12, 135, 82, 20, $_form_password_tt, $ES_PASSWORD))
	$Form_Query = GUICtrlCreateTabItem($_form_tab_query)
	Handle("URL_Query", DialogInput("", 114, 72, 500, 25, $_form_serverurl, 12, 75, 100, 20, $_form_serverurl_tt_query))
	Handle("Filename", DialogInput("", 114, 107, 468, 24, $_form_filename, 12, 110, 100, 20, $_form_filename_tt))
	$Form_SelectFilename = DialogButtonWithIcon($_form_selectfile, 589, 107, 24, 24, $_form_selectfile_tt)
	Handle("Find", DialogInput("", 114, 132, 500, 24, $_form_find, 12, 135, 100, 20, $_form_find_tt))
	Handle("ReplaceWith", DialogInput("", 114, 157, 500, 24, $_form_replacewith, 12, 160, 101, 20, $_form_replacewith_tt))
	Handle("Append", DialogCheckbox($_form_append, 114, 187, 300, 17))
	Handle("Execute", DialogInput("", 114, 212, 468, 24, $_form_execute, 12, 215, 101, 20, $_form_execute_tt))
	$Form_SelectExecute = DialogButtonWithIcon($_form_selectfile, 589, 212, 24, 24, $_form_selectfile_tt)
	UpdateDialog($Form_Profile, $Form_Query, $Form_Synch)
	GUISetState(@SW_SHOW)
	Dim $Loop, $nMsg, $result
	$result = ""
	$Loop = True
	While $Loop
		Switch GUIGetMsg()
			Case $GUI_EVENT_CLOSE
				$Loop = False
			Case $Form_Submit
				UpdateSettings()
				Switch GUICtrlRead(Handle("Tab"))
					Case 0
						$result = QueryServer(Setting("URL_Sync") & "/httpdns.php?hostname=" & Setting("Hostname") & "&password=" & Setting("Password"), "sync")
					Case 1
						$result = QueryServer(Setting("URL_Query"), "query")
					Case Else
						ErrorMessage($_error_unhandledtab & " '" & GUICtrlRead(Handle("Tab")) & "'.")
				EndSwitch
				If $result Then
					$Loop = False
				EndIf
			Case $Form_Abort
				$Loop = False
			Case $Form_SaveProfile
				UpdateSettings()
				SaveSettings(GUICtrlRead($Form_Profile))
				UpdateDialog($Form_Profile, $Form_Query, $Form_Synch)
			Case $Form_Profile
				RestoreSettings(GUICtrlRead($Form_Profile))
				UpdateDialog($Form_Profile, $Form_Query, $Form_Synch)
			Case $Form_SelectFilename
				GUICtrlSetData(Handle("Filename"), FileOpenDialog($_fodlg_filename, "", $_fodlg_filename_pattern, 8, GUICtrlRead(Handle("Filename"))))
			Case $Form_SelectExecute
				GUICtrlSetData(Handle("Execute"), FileOpenDialog($_fodlg_execute, "", $_fodlg_execute_pattern, 8, GUICtrlRead(Handle("Execute"))))
		EndSwitch
	WEnd
	Return $result
EndFunc

; ---------------------
; File Access Routines

Func PatchFileExisting($pattern, $replacement)
	Dim $result, $file_in, $file_out
	$result = False
	$file_in = FileOpen(Setting("Filename"), 0)
	If $file_in <> -1 Then
		$file_out = FileOpen($TempFilename, 2)
		If $file_out <> -1 Then
			Dim $line_in, $line_out, $Loop, $count, $match
			$count = 0
			$Loop = True
			While $Loop
				$line_in = FileReadLine($file_in)
				If Not @error Then
					$match = StringRegExp($line_in, $pattern, 3)
					If Not @error Then
						$count = $count + UBound ($match)
						$line_out = StringRegExpReplace($line_in, $pattern, $replacement)
					Else
						$line_out = $line_in
					EndIf
					If Not FileWriteLine($file_out, $line_out) Then
						ErrorMessage($_error_cannotwritetemp & " '" & $TempFilename & "'.")
						ExitLoop
					EndIf
				Else
					$Loop = False
				EndIf
			WEnd
			FileClose($file_in)
			FileClose($file_out)
			If Not $Loop Then
				If $count > 0 Then
					If FileMove($TempFilename, Setting("Filename"), 1) Then
						$result = True
					Else
						ErrorMessage($_error_cannotmovetemp)
					EndIf
				Else
					If Number(Setting("Append")) = $GUI_CHECKED Then
						$file_out = FileOpen(Setting("Filename"), 1)
						If $file_out <> -1 Then
							FileWriteLine($file_out, $replacement)
							FileClose($file_out)
							$result = True
						Else
							ErrorMessage($_error_cannotopenfile & " '" & Setting("Filename") & "'.")
						EndIf
					Else
						ErrorMessage($_error_nopatternmatch)
					EndIf
				EndIf
			EndIf
			If FileExists($TempFilename) Then
				If Not FileDelete($TempFilename) Then
					ErrorMessage($_error_cannotremovefile & " '" & $TempFilename & "'.")
				EndIf
			EndIf
		Else
			ErrorMessage($_error_cannotopenfile & " '" & $TempFilename & "'.")
		EndIf
	Else
		ErrorMessage($_error_cannotopenfile & " '" & Setting("Filename") & "'.")
	EndIf
	Return $result
EndFunc

Func PatchFile($ip)
	Dim $pattern, $replacement, $result, $file_in, $file_out, $count
	$result = False
	$pattern = StringReplace(Setting("Find"), $PatternIP, $PatternIPAddress)
	$replacement = StringReplace(Setting("ReplaceWith"), $PatternIP, $ip)
	$replacement = StringReplace($replacement, $PatternDate, @YEAR & "-" & @MON & "-" & @MDAY)
	$replacement = StringReplace($replacement, $PatternTime, @HOUR & ":" & @MIN & ":" & @SEC)
	StringRegExp("test", $pattern)
	If Not @error Then
		If FileExists(Setting("Filename")) Then
			$result = PatchFileExisting($pattern, $replacement)
		Else
			If Number(Setting("Append")) = $GUI_CHECKED Then
				$file_out = FileOpen(Setting("Filename"), 9)
				If $file_out <> -1 Then
					FileWriteLine($file_out, $replacement)
					FileClose($file_out)
					$result = True
				Else
					ErrorMessage($_error_cannotopenfile & " '" & Setting("Filename") & "'.")
				EndIf
			Else
				ErrorMessage($_error_filenotfound & " '" & Setting("Filename") & "'.")
			EndIf
		EndIf
	Else
		ErrorMessage($_error_badpattern & " '" & $pattern & "'.")
	EndIf
	Return $result
EndFunc

; -------------------
; Server IO Routines

Func QueryServerParseResponse($response)
	Dim $result, $data, $ip
	$result = ""
	$response = StringStripWS(StringReplace(StringReplace($response, @CR, ""), @LF, ""), 7)
	$response = StringRegExp($response, ".*<body>(.*)</body>.*", 1) ; crop html body
	$response = StringRegExp($response[0], "(?:<[^>]+>)*([^<>]+)(?:<[^>]+>)*", 1) ; wipe all html statements left
	$data = StringRegExp($response[0], "(httpDNS Server) \((PHP)\) Version ([[:digit:]]+.[[:digit:]]+.[[:digit:]]+) : (.*)", 1)
	If Not @error Then
		If $data[3] <> "ACCESS DENIED."  Then
			If Not StringRegExp($data[3], ".*ERROR:.*") Then
				$ip = StringRegExp($data[3], $PatternIPAddress, 1)
				If Not @error Then
					$result = $ip[0]
				Else
					ErrorMessage($_error_invalidip & " '" & $data[3] & "'")
				EndIf
			Else
				ErrorMessage($_error_servererror & " '" & $data[3] & "'")
			EndIf
		Else
			ErrorMessage($_error_invalidpassword)
		EndIf
	Else
		$ip = StringRegExp($response[0], $PatternIPAddress, 1)
		If Not @error Then
			$result = $ip[0]
		Else
			If $response[0] = "N/A"  Then
				ErrorMessage($_error_noipdata)
			Else
				ErrorMessage($_error_invalidresponse & " '" & $response[0] & "'")
			EndIf
		EndIf
	EndIf
	Return $result
EndFunc

Func QueryServer($URL, $command, $mute = Default)
	Dim $result
	$result = False
	If $URL <> "" Then
		Dim $ip
		$ip = QueryServerParseResponse(_INetGetSource($URL))
		If $ip <> "" Then
			Switch $command
				Case "sync" 
					If $mute = Default Or $mute = False Then
						MsgBox(64, $version, $_status_synched & " '" & $ip & "'")
					EndIf
					$result = True
				Case "query" 
					If Setting("Filename") <> "" Then
						$result = PatchFile($ip)
					Else
						$result = True
					EndIf
					If $result And Setting("Execute") <> "" Then
						Dim $execute
						$execute = StringReplace(Setting("Execute"), $PatternIP, $ip)
						Run($execute)
					EndIf
				Case Else
					ErrorMessage($_error_invalidintcommand & " '" & $command & "'")
			EndSwitch
		EndIf
	Else
		ErrorMessage($_error_noserverurl)
	EndIf
	Return $result
EndFunc

; --------------------
; Command Line Parser

Func ParseCommandLine()
	Dim $result
	$result = False
	Switch $CmdLine[0]
		Case 0
			RestoreSettings()
			$result = Dialog(ReadProfileDefault())
		Case 1
			Switch StringLower($CmdLine[1])
				Case "-h", "--help", "/h", "/?" 
					MsgBox(64, $version, $_usage)
					$result = True
				Case Else
					RestoreSettings($CmdLine[1])
					$result = Dialog($CmdLine[1])
			EndSwitch
		Case Else
			Dim $profiles, $profile
			$profiles = Opt("GUIDataSeparatorChar") & ReadProfiles()
			$profiles = StringLeft($profiles, StringLen($profiles) - StringLen($_profile_new))
			$profile = StringRegExp($profiles, ".*\" & Opt("GUIDataSeparatorChar") & "((?i)" & $CmdLine[2] & ")\" & Opt("GUIDataSeparatorChar") & ".*", 1)
			If Not @error Then
				RestoreSettings($profile[0])
				Switch StringLower($CmdLine[1])
					Case "-s", "--sync", "/s" 
						$result = QueryServer(Setting("URL_Sync") & "/httpdns.php?hostname=" & Setting("Hostname") & "&password=" & Setting("Password"), "sync", True)
					Case "-q", "--query", "/q" 
						$result = QueryServer(Setting("URL_Query"), "query", True)
					Case Else
						ErrorMessage($_error_unknowncommand & " '" & $CmdLine[1] & "'")
				EndSwitch
			Else
				ErrorMessage($_error_unknownprofile & " '" & $CmdLine[2] & "'")
			EndIf
	EndSwitch
	Return $result
EndFunc

; -----------
;    MAIN

Dim $errorlevel
$errorlevel = 1
If InitLanguage(@OSLang) Then
	If @OSTYPE = "WIN32_NT"  Then
		If ParseCommandLine() Then
			$errorlevel = 0
		EndIf
	Else
		ErrorMessage($_error_osversion)
	EndIf
EndIf
Exit ($errorlevel)